import pandas as pd
import geopandas as gpd
import osmnx as ox
from h3 import h3
from tqdm import tqdm
from urbanpy.utils import geo_boundary_to_polygon

__all__ = [
    'merge_geom_downloads',
    'filter_population',
    'remove_features',
    'gen_hexagons',
    'merge_shape_hex',
    'overlay_polygons_hexs',
    'resolution_downsampling',
    'osmnx_coefficient_computation',
]

def merge_geom_downloads(gdfs):
    '''
    Merge several GeoDataFrames from OSM download_osm

    Parameters
    ----------

    dfs : array_like
             Array of GeoDataFrames to merge

    Returns
    -------

    concat : GeoDataFrame
                Output from concatenation and unary union of geometries, providing
                a single geometry database for the city

    Examples
    --------

    >>> lima = download_osm(2, "Lima, Peru")
    >>> callao = download_osm(1, "Callao, Peru")
    >>> lima = merge_geom_downloads([lima, callao])
    >>> lima.head()
    geometry
    MULTIPOLYGON (((-76.80277 -12.47562, -76.80261...)))

    '''

    concat = gpd.GeoDataFrame(geometry=[pd.concat(gdfs).unary_union])
    return concat

def filter_population(pop_df, polygon_gdf):
    '''
    Filter an HDX database download to the polygon bounds

    Parameters
    ----------

    pop_df : DataFrame
                Result from download_hdx

    polygon_gdf : GeoDataFrame
                     Result from download_osm or merge_geom_downloads

    Returns
    -------

    filtered_points_gdf : GeoDataFrame
                         Population DataFrame filtered to polygon bounds

    Examples
    --------

    >>> lima = download_osm(2, 'Lima, Peru')
    >>> callao = download_osm(1, 'Callao, Peru')
    >>> lima = merge_geom_downloads([lima, callao])
    >>> pop = download_hdx_population_data("4e74db39-87f1-4383-9255-eaf8ebceb0c9/resource/317f1c39-8417-4bde-a076-99bd37feefce/download/population_per_2018-10-01.csv.zip")
    >>> filter_population(pop, lima)
    latitude   | longitude  | population_2015 | population_2020 | geometry
    -12.519861 | -76.774583 | 2.633668        | 2.644757        | POINT (-76.77458 -12.51986)
    -12.519861 | -76.745972 | 2.633668        | 2.644757        | POINT (-76.74597 -12.51986)
    -12.519861 | -76.745694 | 2.633668        | 2.644757        | POINT (-76.74569 -12.51986)
    -12.519861 | -76.742639 | 2.633668        | 2.644757        | POINT (-76.74264 -12.51986)
    -12.519861 | -76.741250 | 2.633668        | 2.644757        | POINT (-76.74125 -12.51986)

    '''

    minx, miny, maxx, maxy = polygon_gdf.geometry.total_bounds
    limits_filter = pop_df['longitude'].between(minx, maxx) & pop_df['latitude'].between(miny, maxy)
    filtered_points = pop_df[limits_filter]

    geometry_ = gpd.points_from_xy(filtered_points['longitude'], filtered_points['latitude'])
    filtered_points_gdf = gpd.GeoDataFrame(filtered_points, geometry=geometry_, crs='EPSG:4326')

    return filtered_points_gdf

def remove_features(gdf, bounds):
    '''
    Remove a set of features based on bounds

    Parameters
    ----------

    gdf : GeoDataFrame
             Input GeoDataFrame containing the point features filtered with filter_population

    bounds : array_like
                Array input following [minx, miny, maxx, maxy] for filtering (GeoPandas total_bounds method output)


    Returns
    -------

    gdf : GeoDataFrame
             Input DataFrame but without the desired features

    Examples
    --------

    >>> lima = filter_population(pop_lima, poly_lima)
    >>> removed = remove_features(lima, [-12.2,-12, -77.2,-77.17]) #Remove San Lorenzo Island
    >>> print(lima.shape, removed.shape)
    (348434, 4) (348427, 4)

    '''
    minx, miny, maxx, maxy = bounds
    drop_ix = gdf.cx[minx:maxx, miny:maxy].index

    return gdf.drop(drop_ix)

def gen_hexagons(resolution, city):
    '''
    Converts an input multipolygon layer to H3 hexagons given a resolution.

    Parameters
    ----------

    resolution : int, 0:15
                    Hexagon resolution, higher values create smaller hexagons.

    city : GeoDataFrame
              Input city polygons to transform into hexagons.

    Returns
    -------

    city_hexagons : GeoDataFrame
                       Hexagon geometry GeoDataFrame (hex_id, geom).

    Examples
    --------

    >>> lima = filter_population(pop_lima, poly_lima)
    >>> lima_hex = gen_hexagons(8, lima)
    hex	            | geometry
    888e620e41fffff | POLYGON ((-76.80007 -12.46917, -76.80439 -12.4...))
    888e62c809fffff | POLYGON ((-77.22539 -12.08663, -77.22971 -12.0...))
    888e62c851fffff | POLYGON ((-77.20708 -12.08484, -77.21140 -12.0...))
    888e62c841fffff | POLYGON ((-77.22689 -12.07104, -77.23122 -12.0...))
    888e62c847fffff | POLYGON ((-77.23072 -12.07929, -77.23504 -12.0...))

    '''

    # Polyfill the city boundaries
    h3_polygons = list()
    h3_indexes = list()

    # Get every polygon in Multipolygon shape
    city_poly = city.explode().reset_index(drop=True)

    for _, geo in city_poly.iterrows():
        hexagons = h3.polyfill(geo['geometry'].__geo_interface__, res=resolution, geo_json_conformant=True)
        for hexagon in hexagons:
            h3_polygons.append(geo_boundary_to_polygon(hexagon))
            h3_indexes.append(hexagon)

    # Create hexagon dataframe
    city_hexagons = gpd.GeoDataFrame(h3_indexes, geometry=h3_polygons).drop_duplicates()
    city_hexagons.crs = 'EPSG:4326'
    city_hexagons = city_hexagons.rename(columns={0: 'hex'}) # Format column name for readability

    return city_hexagons

def merge_shape_hex(hexs, shape, agg, how='inner', op='intersects'):
    '''
    Merges a H3 hexagon GeoDataFrame with a Point GeoDataFrame and aggregates the
    point gdf data.

    Parameters
    ----------

    hexs : GeoDataFrame
             Input GeoDataFrame containing hexagon geometries

    shape : GeoDataFrame
                Input GeoDataFrame containing points and features to be aggregated

    agg : dict. A dictionary with column names as keys and values as aggregation
             operations. The aggregation must be one of {'sum', 'min', 'max'}.

    how : str. One of {'inner', 'left', 'right'}. Default 'inner'.
                Determines how to merge data:
             'left' uses keys from left and only retains geometry from left
             'right' uses keys from right and only retains geometry from right
             'inner': use intersection of keys from both dfs; retain only left geometry column

    op : str. One of {'intersects', 'contains', 'within'}. Default 'intersects'
                Determines how geometries are queried for merging.

    Returns
    -------

    hexs : GeoDataFrame
                   Result of a spatial join within hex and points. All features are aggregated
                   based on the input parameters

    Examples
    --------

    >>> lima = download_osm(2, 'Lima, Peru')
    >>> pop_lima = download_hdx(...)
    >>> pop_df = filter_population(pop_lima, lima)
    >>> hexs = gen_hexagons(8, lima)
    >>> merge_point_hex(hexs, pop_df, 'inner', 'within', {'population_2020':'sum'})
    0               | geometry                                          | population_2020
    888e628d8bfffff | POLYGON ((-76.66002 -12.20371, -76.66433 -12.2... | NaN
    888e62c5ddfffff | POLYGON ((-76.94564 -12.16138, -76.94996 -12.1... | 14528.039097
    888e62132bfffff | POLYGON ((-76.84736 -12.17523, -76.85167 -12.1... | 608.312696
    888e628debfffff | POLYGON ((-76.67982 -12.18998, -76.68413 -12.1... | NaN
    888e6299b3fffff | POLYGON ((-76.78876 -11.97286, -76.79307 -11.9... | 3225.658803
    
    '''

    joined = gpd.sjoin(shape, hexs, how=how, op=op)

    #Uses index right based on the order of points and hex. Right takes hex index
    hex_merge = joined.groupby('index_right').agg(agg)

    #Avoid SpecificationError by copying the DataFrame
    ret_hex = hexs.copy()

    for key in agg.keys():
        ret_hex.loc[hex_merge.index, key] = hex_merge[key].values

    return ret_hex

def overlay_polygons_hexs(polygons, hexs, hex_col, columns):
    '''
    Overlays a Polygon GeoDataFrame with a H3 hexagon GeoDataFrame and divide the 'columns' the values proportionally to the overlayed area.

    Parameters
    ----------
    polygons : GeoDataFrame
                Input GeoDataFrame containing polygons and columns to be processed

    hexs : GeoDataFrame
             Input GeoDataFrame containing desired output hexagon resolution geometries

    hex_col : str
            Determines the column with the hex id.

    columns : list. A list with column names of the columns that are going to be proportionally adjusted

    Returns
    -------

    hexs : GeoDataFrame
                   Result of a spatial join within hex and points. All columns are adjusted
                   based on the overlayed area.

    Examples
    --------

    >>> overlay_polygons_hexs(zonas_pob, hex_lima, 'hex', pob_vulnerable)
                   hex   POB_TOTAL   geometry
    0  898e6200493ffff  193.705376   POLYGON ((-76.80695 -12.35199, -76.80812 -12.3...
    1  898e6200497ffff  175.749780   POLYGON ((-76.80412 -12.35395, -76.80528 -12.3...
    2  898e620049bffff   32.231078   POLYGON ((-76.81011 -12.35342, -76.81127 -12.3...
    3  898e62004a7ffff   74.154973   POLYGON ((-76.79911 -12.36468, -76.80027 -12.3...
    4  898e62004b7ffff   46.989828   POLYGON ((-76.79879 -12.36128, -76.79995 -12.3...
    '''
    polygons_ = polygons.copy() # Preserve data state
    polygons_['poly_area'] = polygons_.geometry.area # Calc polygon area

    # Overlay intersection
    overlayed = gpd.overlay(polygons_, hexs, how='intersection')

    # Downsample indicators using proporional overlayed area w.r.t polygon area
    area_prop = overlayed.geometry.area / overlayed['poly_area']
    overlayed[columns] = overlayed[columns].apply(lambda col: col * area_prop)

    # Aggregate over Hex ID
    per_hexagon_data = overlayed.groupby(hex_col)[columns].sum()

    # Preserve data as GeoDataFrame
    hex_df = pd.merge(left=per_hexagon_data, right=hexs[[hex_col,'geometry']], on=hex_col)
    hex_gdf = gpd.GeoDataFrame(hex_df[[hex_col]+columns], geometry=hex_df['geometry'], crs=hexs.crs)

    return hex_gdf

def resolution_downsampling(gdf, hex_col, coarse_resolution, agg):
    '''
    Downsample hexagon resolution aggregating indicated metrics (e.g. Transform hexagon resolution from 9 to 6).

    Parameters
    ----------

    gdf : GeoDataFrame
            GeoDataFrame with hexagon geometries (output from gen_hexagons).

    hex_col : str
            Determines the column with the hex id.

    coarse_resolution : int, 0:15
            Hexagon resolution lower than gdf actual resolution (higher values create smaller hexagons).

    Returns
    -------

    gdfc : GeoDataFrame
            GeoDataFrame with lower resolution hexagons geometry and metrics aggregated as indicated.

    '''
    gdf_coarse = gdf.copy()
    coarse_hex_col = 'hex_{}'.format(coarse_resolution)
    gdf_coarse[coarse_hex_col] = gdf_coarse[hex_col].apply(lambda x: h3.h3_to_parent(x,coarse_resolution))
    dfc = gdf_coarse.groupby([coarse_hex_col]).agg(agg).reset_index()
    gdfc_geometry = dfc[coarse_hex_col].apply(geo_boundary_to_polygon)

    return gpd.GeoDataFrame(dfc, geometry=gdfc_geometry, crs=gdf.crs)

def osmnx_coefficient_computation(gdf, net_type, basic_stats, extended_stats, connectivity=False, anc=False, ecc=False, bc=False, cc=False):
    '''
    Apply osmnx's graph from polygon to query a city's street network within a geometry.

    Parameters
    ----------

    gdf : GeoDataFrame
             GeoDataFrame with geometries to download graphs contained within them.

    net_type: str
              Network type to download. One of {'drive', 'drive_service', 'walk', 'bike', 'all', 'all_private'}

    basic_stats : list
                     List of basic stats to compute from downloaded graph

    extended_stats : list
                        List of extended stats to compute from graph

    connectivity : bool. Default False.
                      Compute node and edge connectivity

    anc : bool. Default False.
             Compute avg node connectivity
    ecc : bool. Default False.
             Compute shortest paths, eccentricity and topological metric
    bc : bool. Default False.
             Compute node betweeness centrality
    cc : bool. Default False.
             Compute node closeness centrality

    For more detail about these parameters, see https://osmnx.readthedocs.io/en/stable/osmnx.html#module-osmnx.stats

    Returns
    -------

    gdf : Input GeoDataFrame with updated columns containing the selected metrics

    '''

    #May be a lengthy download depending on the amount of features
    for index, row in tqdm(gdf.iterrows()):
        try:
            graph = ox.graph_from_polygon(row['geometry'], net_type)
            b_stats = ox.basic_stats(graph)
            ext_stats = ox.extended_stats(graph, connectivity, anc, ecc, bc, cc)

            for stat in basic_stats:
                gdf.loc[index, stat] = b_stats.get(stat)
            for stat in extended_stats:
                gdf.loc[index, stat] = ext_stats.get(stat)
        except Exception as err:
            print(f'On record {index}: ', err)

    return gdf



# import pydeck as pdk
import pandas as pd
import plotly.express as px
import matplotlib.pyplot as plt
# from urbanpy.utils import tuples_to_lists

__all__ = [
    # 'pydeck_df',
    # 'gen_pydeck_layer',
    'choropleth_map',
]


# def gen_pydeck_layer(layer_type, data, **kwargs):
#     if layer_type == 'H3HexagonLayer':
#         return pdk.Layer("H3HexagonLayer", data, **kwargs)
#     else:
#         return pdk.Layer('PolygonLayer', data, **kwargs)

def choropleth_map(gdf, color_column, df_filter=None, **kwargs):
    '''
    Produce a Choroplethmap using plotly by passing a GeoDataFrame.

    Parameters
    ----------

    gdf : GeoDataFrame
             Input data containing a geometry column and features

    color_column : str
                      Column from gdf to use as color


    df_filter : pd.Series, default to None
                   Pandas Series containing true/false values that satisfy a
                   condition (e.g. (df['population'] > 100))

    **kargs : Any parameter of plotly.px.choroplethmapbox.

    Examples
    --------

    >>> hex_lima = gen_hexagons(...)
    >>> hex_lima['pop_2020'] = population_2020
    >>> choropleth_map(hex_lima, 'pop_2020', [-12, -77])

    '''


    if df_filter is not None:
        gdff = gdf[df_filter].copy()
    else:
        gdff = gdf.copy()

    gdff = gdff.reset_index()[['index', color_column, 'geometry']].dropna()
    lon, lat = gdff.geometry.unary_union.centroid.xy

    fig = px.choropleth_mapbox(gdff, geojson=gdff[['geometry']].__geo_interface__,
                               color=color_column, locations="index",
                               center={"lat": lat[0], "lon": lon[0]},
                               mapbox_style="carto-positron", **kwargs)
    return fig

# def pydeck_df(gdf, features, cmap, bins, color_feature):
#     '''
#     Prepare a DataFrame for Polygon plotting in PyDeck
#
#     Parameters
#     ----------
#
#     gdf : GeoDataFrame with the geometries to plot
#
#     features : list
#                   List of features to add to the polygon df
#
#     cmap : str
#               Matplotlib colormap to use for plotting
#
#     bins : list
#               Bins to aggregate data into
#
#     color_feature : str
#                        Column name for data to transform into bins and color features
#
#     Returns
#     -------
#
#     polygon_df : DataFrame
#                     df with the coordinates in a list as a column and the selected features
#
#     '''
#
#     cmap = plt.get_cmap(name='magma')
#     json = gdf.__geo_interface__
#
#     json_ = tuples_to_lists(json)
#
#     del json_['bbox']
#
#     df = pd.DataFrame.from_dict(json_)
#
#     polygon_df = pd.DataFrame()
#     polygon_df['coordinates'] = df['features'].apply(lambda row: row['geometry']['coordinates'])
#
#     for feature in features:
#         polygon_df[feature] = df['features'].apply(lambda row: row['properties'][feature])
#
#     polygon_df['bins'] = pd.cut(
#         polygon_df[color_feature],
#         bins = bins
#     )
#
#     bins_labels = polygon_df['bins'].unique().categories.values
#     bins_replace = {label: i  for i, label in enumerate(bins_labels)}
#     polygon_df['count'] = polygon_df['bins'].replace(bins_replace)
#
#     cmap_ixs = [int(round(255 / (i+1))) for i in polygon_df['count'].fillna(0).values]
#     rgb_values = [[color * 255 for color in cmap.colors[cmap_ix]] for cmap_ix in cmap_ixs]
#     rgb_count_df = pd.DataFrame(rgb_values, columns=['r','g','b'])
#
#     polygon_df = pd.concat((polygon_df, rgb_count_df), axis=1)
#
#     polygon_df['r'] = polygon_df['r'].round(0)
#     polygon_df['g'] = polygon_df['g'].round(0)
#     polygon_df['b'] = polygon_df['b'].round(0)
#
#     return polygon_df



import time
import subprocess
import sys
import requests
import googlemaps
import numpy as np
import networkx as nx
import geopandas as gpd
from tqdm.auto import tqdm
# from numba import jit

__all__ = [
    'start_osrm_server',
    'stop_osrm_server',
    'osrm_route',
    'google_maps_dist_matrix',
    'ors_api',
    'compute_osrm_dist_matrix',
    'google_maps_dir_matrix',
    'nx_route'
]

CONTAINER_NAME = 'osrm_routing_server'

def check_container_is_running(container_name):
    '''
    Checks if a container is running

    Parameters
    ----------

    container_name : str
        Name of container to check

    Returns
    -------

    container_running : bool
        True if container is running, False otherwise.

    '''
    completed_process = subprocess.run(['docker', 'ps'], check=True, capture_output=True)
    stdout_str = completed_process.stdout.decode('utf-8')
    container_running = container_name in stdout_str

    return container_running

def start_osrm_server(country, continent):
    '''
    Download data for OSRM, process it and start local osrm server

    Parameters
    ----------

    country : str
             Which country to download data from. Expected in lower case & dashes replace spaces.
    continent : str
             Which continent of the given country. Expected in lower case & dashes replace spaces.

    Examples
    --------

    >>> up.routing.start_osrm_server('peru', 'south-america')
    Starting server ...
    Server was started succesfully.

    '''

    # Download, process and run server command sequence
    dwn_str = f'''
    docker pull osrm/osrm-backend;
    mkdir -p ~/data/osrm/;
    cd ~/data/osrm/;
    wget https://download.geofabrik.de/{continent}/{country}-latest.osm.pbf;
    docker run -t --name osrm_extract -v $(pwd):/data osrm/osrm-backend osrm-extract -p /opt/foot.lua /data/{country}-latest.osm.pbf;
    docker run -t --name osrm_partition -v $(pwd):/data osrm/osrm-backend osrm-partition /data/{country}-latest.osm.pbf;
    docker run -t --name osrm_customize -v $(pwd):/data osrm/osrm-backend osrm-customize /data/{country}-latest.osm.pbf;
    docker container rm osrm_extract osrm_partition osrm_customize;
    docker run -t --name {CONTAINER_NAME}_{continent}_{country} -p 5000:5000 -v $(pwd):/data osrm/osrm-backend osrm-routed --algorithm mld /data/{country}-latest.osm.pbf;
    '''

    container_running = check_container_is_running(CONTAINER_NAME + f"_{continent}_{country}")

    # Check platform
    if sys.platform in ['darwin', 'linux']:
        # Check if container exists:
        if subprocess.run(['docker', 'inspect', CONTAINER_NAME + f"_{continent}_{country}"]).returncode == 0:
            if container_running:
                print('Server is already running.')
            else:
                try:
                    print('Starting server ...')
                    subprocess.run(['docker', 'start', CONTAINER_NAME + f"_{continent}_{country}"], check=True)
                    time.sleep(5) # Wait server to be prepared to receive requests
                    print('Server was started succesfully')
                except subprocess.CalledProcessError as error:
                    print(f'Something went wrong. Please check if port 5000 is being used or check your docker installation.\nError: {error}')

        else:
            try:
                print('This is the first time you used this function.\nInitializing server setup. This may take several minutes...')
                subprocess.Popen(dwn_str, shell=True)

                # Verify container is running
                while container_running == False:
                    container_running = check_container_is_running(CONTAINER_NAME + f"_{continent}_{country}")

                print('Server was started succesfully')
                time.sleep(5) # Wait server to be prepared to receive requests

            except subprocess.CalledProcessError as error:
                print(f'Something went wrong. Please check your docker installation.\nError: {error}')

    elif sys.platform == 'win32':
        print('Still working on windows support')
        #subprocess.Popen(dwn_str.replace(';', '&'))

    else:
        print('Platform not supported')

def stop_osrm_server(country, continent):
    '''
    Run docker stop on the server's container.

        Parameters
        ----------

        country : str
                 Which country osrm to stop. Expected in lower case & dashes replace spaces.
        continent : str
                 Continent of the given country. Expected in lower case & dashes replace spaces.
    '''

    # Check platform
    if sys.platform in ['darwin', 'linux']:
        # Check if container exists:
        if subprocess.run(['docker', 'top', CONTAINER_NAME + f"_{continent}_{country}"]).returncode == 0:
            if check_container_is_running(CONTAINER_NAME + f"_{continent}_{country}") == True:
                try:
                    subprocess.run(['docker', 'stop', CONTAINER_NAME + f"_{continent}_{country}"], check=True)
                    #subprocess.run(['docker', 'container', 'rm', 'osrm_routing_server'])
                    print('Server was stoped succesfully')
                except subprocess.CalledProcessError as error:
                    print(f'Something went wrong. Please check your docker installation.\nError: {error}')
            else:
                print('Server is not running.')

        else:
            print('Server does not exist.')

    elif sys.platform == 'win32':
        print('Still working on windows support')
        #subprocess.Popen(dwn_str.replace(';', '&'))

    else:
        print('Platform not supported')

def osrm_route(origin, destination, profile):
    '''
    Query an OSRM routing server for routes between an origin and a destination
    using a specified profile.

    Parameters
    ----------

    origin : DataFrame with columns x and y or Point geometry
                Input origin in lat lon pairs (y, x) to pass into the routing engine

    destination : DataFrame with columns x and y or Point geometry
                     Input destination in lat lon pairs (y,x) to pass to the routing engine

    profile : str. One of {'foot', 'car', 'bicycle'}
                 Behavior to use when routing and estimating travel time.

    Returns
    -------

    distance : float
                  Total travel distance from origin to destination in meters
    duration : float
                  Total travel time in minutes

    '''
    orig = f'{origin.x},{origin.y}'
    dest = f'{destination.x},{destination.y}'
    url = f'http://localhost:5000/route/v1/{profile}/{orig};{dest}' # Local osrm server
    response = requests.get(url, params={'overview': 'false'})

    try:
        data = response.json()['routes'][0]
        distance, duration = data['distance'], data['duration']
        return distance, duration
    except Exception as err:
        #print(err)
        return np.nan, np.nan

def google_maps_dist_matrix(origin, destination, mode, api_key, **kwargs):
    '''
    Google Maps distance matrix support.

    Parameters
    ----------

    origin : tuple, list or str
            Origin for distance calculation. If tuple or list, a matrix for all
            lat-lon pairs will be computed (origin as rows). If str, google_maps
            will georeference and then compute.

    destination : tuple, list or str
                 Origin for distance calculation. If tuple or list, a matrix for all
                 lat-lon pairs will be computed (origin as rows). If str, google_maps
                 will georeference and then compute.

    mode : str. One of {"driving", "walking", "transit", "bicycling"}
          Mode for travel time calculation

    api_key : str
              Google Maps API key

    **kwargs
        Additional keyword arguments for the distance matrix API. See
        https://github.com/googlemaps/google-maps-services-python/blob/master/googlemaps/distance_matrix.py

    Returns
    -------

    dist : int
           Distance for the o/d pair. Depends on metric parameter

    time : int
           Travel time duration

    Examples
    --------

    >>> API_KEY = 'example-key'
    >>> up.routing.google_maps_dist_matrix('San Juan de Lurigancho', 'Miraflores', 'walking', API_KEY)
        (18477, 13494)
    >>> up.routing.google_maps_dist_matrix((-12,-77), (-12.01,-77.01), 'walking', API_KEY)
        (2428, 1838)
    >>> up.routing.google_maps_dist_matrix([(-12,-77),(-12.11,-77.01)], [(-12.11,-77.01),(-12,-77)], 'walking', API_KEY)
        ([[13743, 0], [0, 13720]], [[10232, 0], [0, 10674]])

    '''

    client = googlemaps.Client(key=api_key)

    try:
        r = client.distance_matrix(origin, destination, mode=mode, **kwargs)

        rows = r['rows']

        dist, dur = [], []

        if len(rows) > 1:
            for row in rows:
                dist.append([element['distance']['value'] for element in row['elements']])
                dur.append([element['duration']['value'] for element in row['elements']])
        else:
            dist = r["rows"][0]["elements"][0]["distance"]["value"]
            dur = r["rows"][0]["elements"][0]["duration"]["value"]
    except Exception as err:
        print(err)
        dist = None
        dur = None

    return dist, dur

def ors_api(locations, origin, destination, profile, metrics, api_key):
    '''
    Interface with OpenRoute Service API for distance matrix computation.

    Parameters
    ----------

    origin : list
            Input origin(s) indices in the location array to compute travel times and distances

    destination : list
            Input destination(s) indices in the location array to compute travel times and distances

    profile : str. One of {'driving-car', 'foot-walking', 'cycling-regular'}

    metrics: list
             Combination of metrics to compute (distance and travel time, or one of them)

    api_key : str
             Authorization API key for OpenRouteService (see API limits)

    Returns
    -------

    dist : list
           Distance matrix

    dur : list
          Travel time matrix

    Examples
    --------

    import requests

    >>> locations = [[9.70093,48.477473],[9.207916,49.153868],[37.573242,55.801281],[115.663757,38.106467]]
    >>> metrics = ["distance","duration"]
    >>> sources = [0]
    >>> destinations = [1,2,3]
    >>> api_key = ...
    >>> up.routing.ors_api(locations, sources, destinations, metrics, 'driving-car', api_key)
    [[5753.86,88998.08,399003.44]]
    [[140861.31,2434228.75,1.0262603E7]]

    '''

    body = {
        "locations": locations,
        "sources": origin,
        "destinations": destination,
        "metrics": metrics
        }

    headers = {
        'Accept': 'application/json, application/geo+json, application/gpx+xml, img/png; charset=utf-8',
        'Authorization': api_key,
        'Content-Type': 'application/json; charset=utf-8'
    }

    r = requests.post(f'https://api.openrouteservice.org/v2/matrix/{profile}', json=body, headers=headers)

    if r.status_code == 200:
        result = r.json()
        return result['distances'], result['durations']
    else:
        return -1, -1

#@jit(forceobj=True)
def compute_osrm_dist_matrix(origins, destinations, profile):
    '''
    Compute distance and travel time origin-destination matrices

    Parameters
    ----------

    origins : GeoDataFrame or GeoSeries
              Input Point geometries corresponding to the starting points of a route.

    destinations : GeoDataFrame or GeoSeries
                   Point geometries corresponding to the end points of a route.

    Returns
    -------

    dist_matrix : array_like
                  Array with origins as rows and destinations as columns. Distance in meters.

    dur_matrix : array_like
                 Array with origins as rows and destinations as columns. Duration in minutes.

    Examples
    --------

    >>> hex, centroids = gen_hexagons(8, lima)
    >>> fs = download_overpass_poi(hex.total_bounds, 'food_supply')
    >>> setup_osrm_server('peru', True)
    >>> dist, dur = compute_osrm_dist_matrix(points.head(), fs.head(), 'walking')
    array([[ 82012.4,  93899.9,  88964.6, 111839.8,  87844. ],
       [ 26346.3,  23725.6,  19618.9,  38607.9,  22725.6],
       [ 26504.4,  23883.7,  19777. ,  38766. ,  22883.7],
       [ 33314.3,  30693.6,  26586.9,  45575.9,  29693.5],
       [ 26659.3,  24038.7,  19931.9,  38920.9,  23038.6]])
    array([[59093.7, 67666.8, 64123.1, 80592.8, 63331.5],
       [19027.5, 17094.6, 14152.3, 27797.6, 16402.8],
       [19141.5, 17208.6, 14266.3, 27911.6, 16516.8],
       [24044.9, 22112. , 19169.7, 32815. , 21420.2],
       [19253.1, 17320.2, 14377.9, 28023.2, 16628.4]])

    '''

    dist_matrix = np.zeros_like([], shape=(origins.shape[0], destinations.shape[0]))
    dur_matrix = np.zeros_like([], shape=(origins.shape[0], destinations.shape[0]))

    if type(origins) == gpd.GeoSeries:
        origins = origins.to_frame()

    if type(destinations) == gpd.GeoSeries:
        destinations = destinations.to_frame()

    for ix, row in tqdm(origins.iterrows(), total=origins.shape[0]):
        for i, r in tqdm(destinations.iterrows(), total=destinations.shape[0]):
            dist, dur = osrm_route(row.geometry, r.geometry, profile)
            dist_matrix[ix, i] = dist
            dur_matrix[ix, i] = dur

    return dist_matrix, dur_matrix

def google_maps_dir_matrix(origin, destination, mode, api_key, **kwargs):
    '''
    Google Maps distance matrix support.

    Parameters
    ----------

    origin : tuple, list or str
            Origin for distance calculation. If tuple or list, a matrix for all
            lat-lon pairs will be computed (origin as rows). If str, google_maps
            will georeference and then compute.

    destination : tuple, list or str
                 Origin for distance calculation. If tuple or list, a matrix for all
                 lat-lon pairs will be computed (origin as rows). If str, google_maps
                 will georeference and then compute.

    mode : str. One of {"driving", "walking", "transit", "bicycling"}
          Mode for travel time calculation

    api_key : str
              Google Maps API key

    **kwargs
        Additional keyword arguments for the distance matrix API. See
        https://github.com/googlemaps/google-maps-services-python/blob/master/googlemaps/directions.py

    Returns
    -------

    dist : int
           Distance for the o/d pair. Depends on metric parameter

    dur : int
           Travel time duration, depends on units kwargs

    Examples
    --------

    >>> API_KEY = 'example-key'
    >>> up.routing.google_maps_dir_matrix('San Juan de Lurigancho', 'Miraflores', 'walking', API_KEY)
        (18477, 13494)

    '''

    client = googlemaps.Client(key=api_key)

    try:
        r = client.directions(origin, destination, mode=mode, **kwargs)

        legs = r[0]['legs']

        dist, dur = 0, 0

        if len(legs) > 1:
            for leg in legs:
                dist += leg['distance']['value']
                dur += leg['duration']['value']
        else:
            dist = legs[0]['distance']['value']
            dur = legs[0]['duration']['value']

    except Exception as err:
        print(err)
        dist = None
        dur = None

    return dist, dur

def nx_route(graph, source, target, weight, length=True):
    '''
    Compute shortest path from a source and target node.

    Parameters
    ----------

    graph : NetworkX Graph
           Input graph from which to calculate paths

    source : str or int
            ID of the source node from which to calculate path. Depending on the graph,
            this may be a string or integer or a combination of both as tuples.

    target : str or int
            ID of the target node. Type corresponds to node id types in the input graph.

    weight : str
            Attribute to be used as weights in the path. If None returns the sequence of nodes
            or the number of nodes to travel as length.

    length : bool
            Flag for whether to calculate the path lenght or the sequence of nodes to follow.
            If weight is none and length is true, the number of nodes in the path will be returned.

    Returns
    -------

    path : list
          Sequence of node ids in a list.

    path_length : float or int
                 Length of the path according to the weight attribute.

    Examples
    --------

    '''
    if length:
        try:
            path_length = nx.shortest_path_length(graph,
                         source,
                         target,
                         weight=weight)
            return path_length
        except:
            #If there is no path within the graph
            return -1
    else:
        try:
            path = nx.shortest_path(graph,
                         source,
                         target,
                         weight=weight)
            return path
        except:
            #If there is no path within the graph
            return -1



import requests
import geopandas as gpd
import pandas as pd
import numpy as np
import osmnx as ox
from shapely.geometry import Polygon

__all__ = [
    'nominatim_osm',
    'hdx_dataset',
    'hdx_fb_population',
    'overpass_pois',
    'osmnx_graph'
]

def nominatim_osm(query, expected_position=0):
    '''
    Download OpenStreetMaps data for a specific city

    Parameters
    ----------

    query: str
        Query for city polygon data to be downloaded

    expected_position: int 0:n
        Expected position of the polygon data within the Nominatim results. Default 0 (first result).

    Returns
    -------

    city: GeoDataFrame
        GeoDataFrame with the city's polygon as its geometry column


    Examples
    --------

    >>> lima = download_osm("Lima, Peru", 2)
    >>> lima.head()
    geometry	 | place_id	 | osm_type	| osm_id     | display_name	| place_rank  |  category | type	       | importance	| icon
    MULTIPOLYGON | 235480647 | relation	| 1944670.0  | Lima, Peru	| 12	      |  boundary |	administrative | 0.703484	| https://nominatim.openstreetmap.org/images/map...

    '''
    osm_url = 'https://nominatim.openstreetmap.org/search.php'
    osm_parameters = {
        'polygon_geojson': '1',
        'format': 'geojson'
    }
    osm_parameters['q'] = query

    response = requests.get(osm_url, params=osm_parameters)
    all_results = response.json()
    gdf = gpd.GeoDataFrame.from_features(all_results['features'])
    city = gdf.iloc[expected_position:expected_position+1, :]

    return city

def hdx_dataset(resource):
    '''
    Download the High Resolution Population Density maps from HDX.

    Parameters
    ----------

    resource : str
                  Specific address to the resource for each city. Since every dataset
                  is referenced to a diferent resource id, only the base url can be provided
                  by the library

    Returns
    -------

    population : DataFrame
                    DataFrame with lat, lon, and population columns. Coordinates
                    are in EPSG 4326.


    Examples
    --------

    >>> pop_lima = download_hdx_population_data("4e74db39-87f1-4383-9255-eaf8ebceb0c9/resource/317f1c39-8417-4bde-a076-99bd37feefce/download/population_per_2018-10-01.csv.zip")
    >>> pop_lima.head()
    latitude   | longitude  | population_2015 |	population_2020
    -18.339306 | -70.382361 | 11.318147	      | 12.099885
    -18.335694 | -70.393750 | 11.318147	      | 12.099885
    -18.335694 | -70.387361	| 11.318147	      | 12.099885
    -18.335417 | -70.394028	| 11.318147	      | 12.099885
    -18.335139 | -70.394306	| 11.318147	      | 12.099885

    '''
    hdx_url = f'https://data.humdata.org/dataset/{resource}'
    population = pd.read_csv(hdx_url)
    return population

def hdx_fb_population(country, map):
    '''
    Download population density maps from Facebook HDX

    Parameters
    ----------

    country: str. One of {'argentina', 'bolivia', 'brazil', 'chile', 'colombia', 'ecuador', 'paraguay', 'peru', 'uruguay'}
            Input country to download data from.

    Returns
    -------

    population: DataFrame
                DataFrame with lat, lon, and population columns. Coordinates
                are in EPSG 4326.

    map: str. One of {'full', 'children', 'youth', 'elderly'}
         Input population map to download

    Examples
    --------

    >>> urbanpy.download.hdx_fb_population('peru', 'full')
    latitude   | longitude  | population_2015 |	population_2020
    -18.339306 | -70.382361 | 11.318147	      | 12.099885
    -18.335694 | -70.393750 | 11.318147	      | 12.099885
    -18.335694 | -70.387361	| 11.318147	      | 12.099885
    -18.335417 | -70.394028	| 11.318147	      | 12.099885
    -18.335139 | -70.394306	| 11.318147	      | 12.099885

    '''

    dataset_dict = {
        'argentina': {
            'full': 'https://data.humdata.org/dataset/6cf49080-1226-4eda-8700-a0093cbdfe4d/resource/5737d87f-e17f-4c82-b1bd-d589ed631318/download/population_arg_2018-10-01.csv.zip',
            'children': 'https://data.humdata.org/dataset/6cf49080-1226-4eda-8700-a0093cbdfe4d/resource/1795ad97-e06a-4ca4-83aa-32ab612f55ba/download/arg_children_under_five_2019-06-01_csv.zip',
            'youth': 'https://data.humdata.org/dataset/6cf49080-1226-4eda-8700-a0093cbdfe4d/resource/dff11b02-a356-4a4f-948c-5a7722ad7365/download/arg_youth_15_24_2019-06-01_csv.zip',
            'elderly':'https://data.humdata.org/dataset/6cf49080-1226-4eda-8700-a0093cbdfe4d/resource/8f1d9473-90f0-441f-b464-c76960e9e130/download/arg_elderly_60_plus_2019-06-01_csv.zip'
        },
        'bolivia': {
            'full': 'https://data.humdata.org/dataset/64f916a6-2f35-4399-8971-25e18fdb09bd/resource/d5fc8980-f3f2-4523-ac4d-f188201518d5/download/population_bol_2018-10-01.csv.zip',
            'children': 'https://data.humdata.org/dataset/64f916a6-2f35-4399-8971-25e18fdb09bd/resource/4fe96e38-3895-4f5e-b3f9-df6317a9752f/download/bol_children_under_five_2019-06-01_csv.zip',
            'youth': 'https://data.humdata.org/dataset/64f916a6-2f35-4399-8971-25e18fdb09bd/resource/7c5fba5b-cdb6-49f2-8ae5-5337734444d2/download/bol_youth_15_24_2019-06-01_csv.zip',
            'elderly': 'https://data.humdata.org/dataset/64f916a6-2f35-4399-8971-25e18fdb09bd/resource/8acccec8-a1a4-42d7-b301-fb734324960e/download/bol_elderly_60_plus_2019-06-01_csv.zip'
        },
        'brazil': {
            'full': [
                'https://data.humdata.org/dataset/c17003d1-47f4-4ec5-8229-2f77aeb114be/resource/957218ee-c740-44c0-88e5-7faeef813a0c/download/population_bra_northeast_2018-10-01.csv.zip',
                'https://data.humdata.org/dataset/c17003d1-47f4-4ec5-8229-2f77aeb114be/resource/1e1f271b-1055-4365-b391-f6fdf3093fe2/download/population_bra_northwest_2018-10-01.csv.zip',
                'https://data.humdata.org/dataset/c17003d1-47f4-4ec5-8229-2f77aeb114be/resource/eb17516f-3c84-4626-95e4-df1f342f3d82/download/population_bra_southeast_2018-10-01.csv.zip',
                'https://data.humdata.org/dataset/c17003d1-47f4-4ec5-8229-2f77aeb114be/resource/5cb55d1a-9f11-4004-82f3-0c27e878495a/download/population_bra_southwest_2018-10-01.csv.zip'
            ],
            'children': 'https://data.humdata.org/dataset/c17003d1-47f4-4ec5-8229-2f77aeb114be/resource/54964df0-8d6a-4f65-ac10-bcf11499a9fe/download/bra_children_under_five_2019-06-01_csv.zip',
            'youth': 'https://data.humdata.org/dataset/c17003d1-47f4-4ec5-8229-2f77aeb114be/resource/0e4af4ff-e4e5-4686-a17c-c8b326d98d76/download/bra_youth_15_24_2019-06-01_csv.zip',
            'elderly':'https://data.humdata.org/dataset/c17003d1-47f4-4ec5-8229-2f77aeb114be/resource/6d6f1ea3-c8d3-4dc0-b0ea-af9c92b14b48/download/bra_elderly_60_plus_2019-06-01_csv.zip'
        },
        'chile': {
            'full': 'https://data.humdata.org/dataset/dd47e052-02cc-4a3f-972a-421d600b3d85/resource/bb560451-9c50-4d57-8ff3-872fa260c102/download/population_chl_2018-10-01.csv.zip',
            'children': 'https://data.humdata.org/dataset/dd47e052-02cc-4a3f-972a-421d600b3d85/resource/33b43b9a-9f47-4f25-8bdc-568e8850fde8/download/chl_children_under_five_2019-06-01_csv.zip',
            'youth': 'https://data.humdata.org/dataset/dd47e052-02cc-4a3f-972a-421d600b3d85/resource/2c4083b5-e682-4ae5-8df8-b2934c4eef9c/download/chl_youth_15_24_2019-06-01_csv.zip',
            'elderly': 'https://data.humdata.org/dataset/dd47e052-02cc-4a3f-972a-421d600b3d85/resource/056a5832-1490-41e4-8ee1-7c90ff1389ff/download/chl_elderly_60_plus_2019-06-01_csv.zip'
        },
        'colombia': {
            'full': 'https://data.humdata.org/dataset/2f865527-b7bf-466c-b620-c12b8d07a053/resource/357c91e0-c5fb-4ae2-ad9d-00805e5a075d/download/population_col_2018-10-01.csv.zip',
            'children': 'https://data.humdata.org/dataset/2f865527-b7bf-466c-b620-c12b8d07a053/resource/e8422e0d-0c1a-4aff-b790-76fffa8e09a6/download/col_children_under_five_2019-06-01_csv.zip',
            'youth': 'https://data.humdata.org/dataset/2f865527-b7bf-466c-b620-c12b8d07a053/resource/f458dcfc-3441-4bec-b45f-d91081801501/download/col_youth_15_24_2019-06-01_csv.zip',
            'elderly': 'https://data.humdata.org/dataset/2f865527-b7bf-466c-b620-c12b8d07a053/resource/3e871e9d-d9fa-4d52-af47-43cae54c7a6d/download/col_elderly_60_plus_2019-06-01_csv.zip'
        },
        'ecuador': {
            'full': 'https://data.humdata.org/dataset/58c3ac3f-febd-4222-8969-59c0fe0e7a0d/resource/c05a3c81-a78c-4e6c-ac05-de1316d4ba12/download/population_ecu_2018-10-01.csv.zip',
            'children': 'https://data.humdata.org/dataset/58c3ac3f-febd-4222-8969-59c0fe0e7a0d/resource/80e46cf3-1906-41a9-a779-8f501cab48a5/download/ecu_children_under_five_2019-06-01_csv.zip',
            'youth': 'https://data.humdata.org/dataset/58c3ac3f-febd-4222-8969-59c0fe0e7a0d/resource/57362a78-e2fa-4f71-8876-1a6d67e27fe5/download/ecu_youth_15_24_2019-06-01_csv.zip',
            'elderly': 'https://data.humdata.org/dataset/58c3ac3f-febd-4222-8969-59c0fe0e7a0d/resource/904d8988-d18d-41a5-a7f7-22668204cefe/download/ecu_elderly_60_plus_2019-06-01_csv.zip'
        },
        'paraguay': {
            'full': 'https://data.humdata.org/dataset/318c589b-9091-4aa4-b384-351208be71e9/resource/6c21c73c-05ba-4818-9574-cb56fa04b210/download/population_pry_2018-10-01.csv.zip',
            'children': 'https://data.humdata.org/dataset/318c589b-9091-4aa4-b384-351208be71e9/resource/7ff05dd3-61f0-40a5-90ee-7262b4190ac2/download/pry_children_under_five_2019-06-01_csv.zip',
            'youth': 'https://data.humdata.org/dataset/318c589b-9091-4aa4-b384-351208be71e9/resource/226e0efe-73b1-4d4b-96f1-06cb3ffd165a/download/pry_youth_15_24_2019-06-01_csv.zip',
            'elderly': 'https://data.humdata.org/dataset/318c589b-9091-4aa4-b384-351208be71e9/resource/8cb53c32-b8f9-4f6f-9d22-50478040fafc/download/pry_elderly_60_plus_2019-06-01_csv.zip'
        },
        'peru': {
            'full': 'https://data.humdata.org/dataset/4e74db39-87f1-4383-9255-eaf8ebceb0c9/resource/317f1c39-8417-4bde-a076-99bd37feefce/download/population_per_2018-10-01.csv.zip',
            'children': 'https://data.humdata.org/dataset/4e74db39-87f1-4383-9255-eaf8ebceb0c9/resource/ed2712e7-4668-45a2-b76f-eb819df9e0c1/download/per_children_under_five_2019-06-01_csv.zip',
            'youth': 'https://data.humdata.org/dataset/4e74db39-87f1-4383-9255-eaf8ebceb0c9/resource/86daf899-8e63-4262-be92-15934e59cabf/download/per_youth_15_24_2019-06-01_csv.zip',
            'elderly': 'https://data.humdata.org/dataset/4e74db39-87f1-4383-9255-eaf8ebceb0c9/resource/8cc100cf-68a4-4fda-a8e6-a63b99ad5b00/download/per_elderly_60_plus_2019-06-01_csv.zip'
        },
        'uruguay': {
            'full': 'https://data.humdata.org/dataset/61e8075b-5c42-495a-98eb-01b3e819dcb5/resource/69900b71-8d0b-49d5-a235-f9b5cc5b820a/download/population_ury_2018-10-01.csv.zip',
            'children': 'https://data.humdata.org/dataset/61e8075b-5c42-495a-98eb-01b3e819dcb5/resource/90d47ee3-5e5f-4aa5-a999-69b1f3e82061/download/ury_children_under_five_2019-06-01_csv.zip',
            'youth': 'https://data.humdata.org/dataset/61e8075b-5c42-495a-98eb-01b3e819dcb5/resource/8ed64dfe-449c-4026-9b48-0416f67c3f38/download/ury_youth_15_24_2019-06-01_csv.zip',
            'elderly': 'https://data.humdata.org/dataset/61e8075b-5c42-495a-98eb-01b3e819dcb5/resource/b9cfa194-b097-4611-a96e-254014990d8a/download/ury_elderly_60_plus_2019-06-01_csv.zip'
        }
    }

    #Brazil is split into 4 maps
    if type(dataset_dict[country][map]) == list:
        return pd.concat([pd.read_csv(file) for file in dataset_dict[country][map]])
    else:
        return pd.read_csv(dataset_dict[country][map])

def overpass_pois(bounds, facilities=None, custom_query=None):
    '''
    Download POIs using Overpass API

    Parameters
    ----------

    bounds: array_like
                Input bounds for query. Follows [minx,miny,maxx,maxy] pattern.

    facilities: {'food', 'health', 'education', 'finance'}
                Type of facilities to download according to HOTOSM types. Based on this a different type of query is constructed.

    custom_query: str (Optional)
                String with custom Overpass QL query (See https://wiki.openstreetmap.org/wiki/Overpass_API/Language_Guide). If this parameter is diferent than None, bounds and facilities values are ignored. Defaults to None.


    Returns
    -------

    gdf: GeoDataFrame containing all the POIs from the selected type of facility

    response: Only if 'custom_query' is given. Returns an HTTP response from the Overpass Server

    Examples
    --------

    '''
    minx, miny, maxx, maxy = bounds

    bbox_string = f'{minx},{miny},{maxx},{maxy}'

        # Definir consulta para instalaciones de oferta de alimentos en Lima
    overpass_url = "http://overpass-api.de/api/interpreter"

    facilities_opt = {
        'food': 'node["amenity"="marketplace"];\nnode["shop"~"supermarket|kiosk|mall|convenience|butcher|greengrocer"];',
        'health': 'node["amenity"~"doctors|dentist|clinic|hospital|pharmacy"];',
        'education': 'node["amenity"~"kindergarten|school|college|university"];',
        'finance': 'node["amenity"~"mobile_money_agent|bureau_de_change|bank|microfinance|atm|sacco|money_transfer|post_office"];',
    }

    if custom_query == None:
        overpass_query = f"""
            [timeout:120][out:json][bbox];
            (
                 {facilities_opt[facilities]}
            );
            out body geom;
            """
        # Request data
        response = requests.get(overpass_url, params={'data': overpass_query,
                                                      'bbox': bbox_string})
        data = response.json()
        df = pd.DataFrame.from_dict(data['elements'])
        df_geom = gpd.points_from_xy(df['lon'], df['lat'])
        gdf = gpd.GeoDataFrame(df, geometry=df_geom)

        gdf['poi_type'] = gdf['tags'].apply(lambda tag: tag['amenity'] if 'amenity' in tag.keys() else np.NaN)

        if facilities == 'food':
            # Food facilities also have its POI type wthin the shop tag (See query)
            also_poi_type = gdf['tags'].apply(lambda tag: tag['shop'] if 'shop' in tag.keys() else np.NaN)
            gdf['poi_type'] = gdf['poi_type'].fillna(also_poi_type)

        return gdf

    else:
        response = requests.get(overpass_url, params={'data': custom_query, 'bbox': bbox_string})
        return response

def osmnx_graph(download_type, network_type='drive', query_str=None,
                geom=None, distance=None, **kwargs):
    '''
    Download a graph from OSM using osmnx.

    Parameters
    ----------

    download_type: str. One of {'polygon', 'place', 'point'}
                   Input download type. If polygon, the polygon parameter must be
                   provided as a Shapely Polygon.
    network_type: str. One of {'drive', 'drive_service', 'walk', 'bike', 'all', 'all_private'}
                  Network type to download. Defaults to drive.

    query_str: str
               Optional. Only requiered for place type downloads. Query string to download a network.

    polygon: Shapely Polygon or Point
             Optional. Polygon requiered for polygon type downloads, Point for place downloads.
             Polygons are used as bounds for network download, points as the center with a distance buffer.

    distance: int
              Distance in meters to use as buffer from a point to download the network.

    Returns
    -------

    G: networkx MultiDiGraph
       Requested graph with simplyfied geometries

    Examples
    --------

    >>> poly = urbanpy.download.nominatim_osm('San Isidro, Peru')
    >>> G = urbanpy.download.osmnx_graph('polygon', geom=lima.loc[0,'geometry'])
    <networkx.classes.multidigraph.MultiDiGraph at 0x1a2ba08150>

    '''

    if (download_type == 'polygon') and (geom is not None) and (type(geom) == Polygon):
        G = ox.graph_from_polygon(geom)
        return G

    elif (download_type == 'point') and (geom is not None) and (distance is not None):
        G = ox.graph_from_point(geom, dist=distance)
        return G

    elif download_type == 'place' and query_str is not None:
        G = ox.graph_from_place(query_str)
        return G

    elif download_type == 'polygon' and geom is None:
        print('Please provide a polygon to download a network from.')

    elif download_type == 'place' and query_str is None:
        print('Please provide a query string to download a network from.')

    else:
        if distance is None and download_type == 'point':
            print('Please provide a distance buffer for the point download')

        if geom is None and distance is not None:
            print('Please provide a Point geometry.')



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt', 'r') as f:
    install_requires = f.read().split('\n')[:-1]

setuptools.setup(
     name='urbanpy',
     version='0.2',
     author="Andres Regal, Claudio Ortega",
     author_email="a.regalludowieg@up.edu.pe",
     description="A library to download, process and visualize high resolution urban data.",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://github.com/IngenieriaUP/urbanpy",
     packages=setuptools.find_packages(),
     classifiers=[
         "Programming Language :: Python :: 3",
         "Operating System :: OS Independent",
     ],
     install_requires=install_requires,
     python_requires='>=3.6'
 )



#from .urbanpy import *

from . import download, geom, plotting, routing, utils
# 
# __all__ = []
#
# __all__.extend(download.__all__)
# __all__.extend(geom.__all__)
# __all__.extend(plotting.__all__)
# __all__.extend(routing.__all__)
# __all__.extend(utils.__all__)



import pandas as pd
import numpy as np
from h3 import h3
from math import ceil
from shapely.geometry import Polygon
from sklearn.neighbors import BallTree
#from numba import jit

__all__ = [
    'swap_xy',
    'nn_search',
    #'tuples_to_lists',
    'shell_from_geometry',
    'geo_boundary_to_polygon',
    'create_duration_labels'
]

def swap_xy(geom):
    '''
    Util function in case an x,y coordinate needs to be switched

    Parameters
    ----------

    geom : GeoSeries
              Input series containing the geometries needing a coordinate swap

    Returns
    -------

    shell : list
               List containing the exterior borders of the geometry
    holes : list
               Array of all holes within a geometry. Only for Polygon and Multipolygon
    coords : list
                List of geomerty type with the swapped x,y coordinates

    '''
    if geom.is_empty:
        return geom

    if geom.has_z:
        def swap_xy_coords(coords):
            for x, y, z in coords:
                yield (y, x, z)
    else:
        def swap_xy_coords(coords):
            for x, y in coords:
                yield (y, x)

    # Process coordinates from each supported geometry type
    if geom.type in ('Point', 'LineString', 'LinearRing'):
        return type(geom)(list(swap_xy_coords(geom.coords)))
    elif geom.type == 'Polygon':
        ring = geom.exterior
        shell = type(ring)(list(swap_xy_coords(ring.coords)))
        holes = list(geom.interiors)
        for pos, ring in enumerate(holes):
            holes[pos] = type(ring)(list(swap_xy_coords(ring.coords)))
        return type(geom)(shell, holes)
    elif geom.type.startswith('Multi') or geom.type == 'GeometryCollection':
        # Recursive call
        return type(geom)([swap_xy(part) for part in geom.geoms])
    else:
        raise ValueError('Type %r not recognized' % geom.type)

def nn_search(tree_features, query_features, metric='haversine'):
    '''
    Build a BallTree for nearest neighbor search based on selected distance.

    Parameters
    ----------

    tree_features : array_like
                       Input features to create the search tree. Features are in
                       lat, lon format.

    query_features : array_like
                        Points to which calculate the nearest neighbor within the tree.
                        lat, lon coordinates expected.

    metric : str
                Distance metric for neighorhood search. Default haversine for latlon coordinates. If haversine is selected lat, lon coordinates are converted to radians.

    Returns
    -------

    dist : array_like
                   Array with the corresponding distance in km (if haversine then distance * earth radius)

    ind : array_like
                   Array with the corresponding index of the tree_features values.

    '''

    if metric == 'haversine':

        tf = np.radians(tree_features)
        tree = BallTree(tf, metric=metric)

        qf = np.radians(query_features)

        dist, ind = tree.query(qf)
        return dist * 6371000/1000, ind
    else:
        tree = BallTree(tree_features, metric=metric)
        dist, ind = tree.query(query_features)
        return dist, ind

def shell_from_geometry(geometry):
    '''
    Util function for park and pitch processing.
    '''

    shell = []
    for record in geometry:
        shell.append([record['lon'], record['lat']])
    return shell

def geo_boundary_to_polygon(x):
    '''
    Transform h3 geo boundary to shapely Polygon

    Parameters
    ----------

    x : str
        H3 hexagon index

    Returns
    -------

    polygon : Polygon
        Polygon representing H3 hexagon area

    '''
    return Polygon([bound[::-1] for bound in h3.h3_to_geo_boundary(x)]) # format as x,y (lon, lat)

def create_duration_labels(durations):
    '''
    Creates inputs for pd.cut function (bins and labels) especifically for the trip durations columns.

    Parameters
    ----------

    durations : Pandas Series
        Series containing trip durations in minutes.

    Returns
    -------

    custom_bins : list
        List of numbers with the inputs for the bins parameter of pd.cut function

    custom_labels : list
        List of numbers with the inputs for the labels parameter of pd.cut function

    '''
    default_bins = [0, 15, 30, 45, 60, 90, 120]
    default_labels = ["De 0 a 15", "De 15 a 30", "De 30 a 45", "De 45 a 60",
                      "De 60 a 90", "De 90 a 120", "Más de 120"]

    bins_ = default_bins.copy()

    max_duration_raw = durations.max()
    max_duration_asint = ceil(max_duration_raw)

    bins_.insert(0, max_duration_asint)
    bins_ = sorted(set(bins_))
    ix = bins_.index(max_duration_asint)

    if (ix + 1) >= len(default_bins) and max_duration_asint != default_bins[-1]:
        default_bins.append(max_duration_asint)

    custom_bins = default_bins[:ix + 1]
    custom_labels = default_labels[:ix]

    return custom_bins, custom_labels

# @jit(forceobj=True)
# def tuples_to_lists(json):
#     '''
#     Util function to convert the geo interface of a GeoDataFrame to PyDeck GeoJSON format.
#
#     Parameters
#     ----------
#
#     json : GeoJSON from a GeoDataFrame.__geo_interface__
#
#
#     Returns
#     -------
#
#     json : dict
#               GeoJSON with the corrected features
#
#     Examples
#     --------
#
#     >>> json = { 'type': 'FeatureCollection', 'features': [{'id': 0, 'geometry': {'coordinates': (((-77, -12), (-77.1, -12.1)))}}]}
#
#     >>> tuples_to_lists(json)
#
#     {'type': 'FeatureCollection', 'features': [{'id': 0, 'geometry': {'coordinates': [[[-77, -12], [-77.1, -12.1]]]}}]}
#
#     '''
#
#     for i in range(len(json['features'])):
#         t = [list(x) for x in json['features'][i]['geometry']['coordinates']]
#         poly = [[list(x) for x in t[0]]]
#
#         json['features'][i]['geometry']['coordinates'] = poly
#
#     return json
